"""
# Probabilistic graphical models (PGMs)
"""

from __future__ import annotations

import logging
import pickle
import re
from copy import deepcopy
from functools import reduce
from typing import Iterable, TypeVar

import networkx as nx

from .nodes import Node, NodeGroup, transform_parameter
from .types import ModelState, NodeState, TFPBijectorClass

logger = logging.getLogger(__name__)


def _join_args(args: Iterable[str]) -> str:
    args = [arg for arg in args if arg != ""]
    return ", ".join(args)


def _nodes_to_args(nodes: Iterable[Node], short: bool = False) -> str:
    args = [f"{node:s}" for node in nodes]

    if short and len(args) > 3:
        args = args[0:3] + ["..."]

    return _join_args(args)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Model ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


T = TypeVar("T", Node, NodeGroup)


class Model:
    """
    A probabilistic graphical model.

    A model is defined by the input-output relations between a number of
    `liesel.liesel.nodes.Node` objects. This class provides methods to compute
    the log-probability of a model and to update its nodes efficiently.

    **Caveat emptor:** This class does not check for missing inputs or groups
    that are not fully contained in a model. Use the `ModelBuilder` to set up
    a model in a safe way.

    ## Attributes

    - `graph`: The graph of the input-output relations between the nodes in the model.
    - `groups`: A dictionary of the nodes groups in the model with their names as keys.
    - `nodes`: A dictionary of the nodes in the model with their names as keys.
    - `sorted_nodes`: A list of the nodes in the model in a topological order.
    """

    def __init__(
        self,
        nodes: Iterable[Node] | None = None,
        groups: Iterable[NodeGroup] | None = None,
    ) -> None:
        nodes = self._update_names(nodes or [], "n")
        self.nodes = {node.name: node for node in nodes}

        groups = self._update_names(groups or [], "g")
        self.groups = {group.name: group for group in groups}

        self.graph = nx.DiGraph()
        self.graph.add_nodes_from(nodes)

        for node in nodes:
            node.outputs = set()

        for node in nodes:
            for input in node.inputs:
                self.graph.add_edge(input, node)
                input.outputs.add(node)

        self.sorted_nodes = list(nx.topological_sort(self.graph))

        for node in self.sorted_nodes:
            node.model = self
            node.update()

    @staticmethod
    def _update_names(xs: Iterable[T], prefix: str) -> Iterable[T]:
        """
        Sets any missing node or group names during initialization and checks for
        duplicate names.
        """

        names = [x.name for x in xs if x.has_name]
        counter = 0

        for x in xs:
            if not x.has_name:
                while prefix + str(counter) in names:
                    counter += 1

                x.name = prefix + str(counter)
                names.append(x.name)

        if len(names) != len(set(names)):
            _type = "node" if isinstance(next(iter(xs)), Node) else "group"
            raise RuntimeError(f"Duplicate {_type} names supplied to Model()")

        return xs

    def empty_copy(self) -> Model:
        """Returns a deep copy of the model with an empty state."""

        state = self.state.copy()

        zeros = {name: NodeState(0.0, 0.0) for name in self.nodes}

        self.state = zeros
        copy = deepcopy(self)
        self.state = state

        return copy

    def get_nodes_by_class(self, cls: type[Node]) -> dict[str, Node]:
        """Returns the nodes of the provided class from the model."""

        return {
            name: node  # black: break line
            for name, node in self.nodes.items()
            if isinstance(node, cls)
        }

    def get_nodes_by_regex(self, regex: str) -> dict[str, Node]:
        """Returns the nodes with matching names from the model."""

        _regex = re.compile(regex)

        return {
            name: node  # black: break line
            for name, node in self.nodes.items()
            if _regex.search(name)
        }

    @property
    def jaxified(self) -> bool:
        """Whether JAX NumPy is enabled for all nodes in the model."""

        return all(node.jaxified for node in self.sorted_nodes)

    @jaxified.setter
    def jaxified(self, jaxified: bool) -> None:
        if jaxified:
            self.jaxify()
        else:
            self.unjaxify()

    def jaxify(self) -> Model:
        """Enables JAX NumPy for all nodes in the model."""

        for node in self.sorted_nodes:
            node.jaxify()

        return self

    @property
    def log_prob(self) -> float:
        """
        Returns the log-probability of the model.

        The log-probability is computed as the sum of the log-probabilities
        of all nodes in the model. In a Bayesian context, it can be understood
        as the unnormalized log-posterior.
        """

        log_probs = [node.log_prob for node in self.sorted_nodes]
        return reduce(lambda x, y: x + y, log_probs)

    @property
    def state(self) -> ModelState:
        """
        Returns the state of the model.

        A model state is a dictionary of node names and states.
        """

        return {node.name: node.state for node in self.sorted_nodes}

    @state.setter
    def state(self, state: ModelState) -> None:
        for name, node_state in state.items():
            self.nodes[name].state = node_state

    def transform_parameter(self, name: str, bijector: str | TFPBijectorClass) -> Model:
        """
        Returns a deep copy of the model with the transformed parameter.

        The node groups are currently not copied over to the new model.
        """

        nodes = deepcopy(self.nodes)

        group = transform_parameter(nodes[name], bijector)

        old = nodes[name]
        new = group["original"]
        nodes[name] = new

        for output in old.outputs:
            if output.has_calculator:
                inputs = output.calculator.inputs.replace(old, new)
                output.calculator._inputs = inputs

            if output.has_distribution:
                inputs = output.distribution.inputs.replace(old, new)
                output.distribution._inputs = inputs

        mb = ModelBuilder()
        mb.add_nodes(nodes.values())
        return mb.build()

    def unjaxify(self) -> Model:
        """Disables JAX NumPy for all nodes in the model."""

        for node in self.sorted_nodes:
            node.unjaxify()

        return self

    def update(self) -> Model:
        """
        Updates all outdated nodes in the model.

        The update is performed in a topological order, restoring a consistent state
        of the model. This method is called automatically by the nodes if their value
        has changed (unless requested otherwise by the user).
        """

        for node in self.sorted_nodes:
            if node.outdated:
                node.update()

        return self

    def __setstate__(self, state):
        self.__dict__.update(state)

        for node in self.sorted_nodes:
            node.model = self

        self.jaxified = self.jaxified

    def __repr__(self) -> str:
        cls = type(self).__name__
        nodes = self.sorted_nodes

        if nodes:
            args = _nodes_to_args(nodes, short=True)
            return f"{cls}([{args}])"

        return f"{cls}()"


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Model builder ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class ModelBuilder:
    """
    A builder class for the `Model`.

    ## Attributes

    - `nodes`: A list of nodes to be added to the model.
    - `groups`: A list of node groups to be added to the model.
    """

    def __init__(
        self,
        nodes: Iterable[Node] | None = None,
        groups: Iterable[NodeGroup] | None = None,
    ) -> None:
        self.nodes = list(nodes or [])
        self.groups = list(groups or [])

    def add_nodes(self, *nodes: Node | Iterable[Node]) -> ModelBuilder:
        """Adds nodes to the model builder."""

        for arg in nodes:
            if isinstance(arg, Node):
                self.nodes.append(arg)
            else:
                self.add_nodes(*arg)

        return self

    def add_groups(self, *groups: NodeGroup | Iterable[NodeGroup]) -> ModelBuilder:
        """Adds node groups to the model builder."""

        for arg in groups:
            if isinstance(arg, NodeGroup):
                self.groups.append(arg)
            else:
                self.add_groups(*arg)

        return self

    @staticmethod
    def _add_inputs(nodes: Iterable[Node]) -> list[Node]:
        """Adds the inputs to an iterable of nodes."""

        nodes = list(nodes)
        visited = []

        while nodes:
            node = nodes.pop(0)  # pop from the left, maintain order

            if node not in visited:
                nodes.extend(node.inputs)
                visited.append(node)

        return visited

    def all_nodes(self) -> list[Node]:
        """
        Returns a list of all *unique* nodes that will be part of the model,
        including the nodes from the groups and the inputs.
        """

        nodes = self.nodes.copy()

        for group in self.groups:
            nodes.extend(group.values())

        nodes = self._add_inputs(nodes)
        nodes = list(dict.fromkeys(nodes))  # remove duplicates, maintain order
        return nodes

    def build(self) -> Model:
        """Builds the model, including the inputs."""

        nodes = self.all_nodes()

        if not nodes:
            logger.warning(f"No nodes in {repr(self)}, building an empty model")

        return Model(nodes, self.groups)

    def __repr__(self) -> str:
        cls = type(self).__name__
        nodes = self.all_nodes()

        if nodes:
            args = _nodes_to_args(nodes, short=True)
            return f"{cls}([{args}])"

        return f"{cls}()"


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Save and load models ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def save_model(model: Model, path: str) -> None:
    """Saves a model to a pickle file."""

    with open(path, "wb") as handle:
        pickle.dump(model, handle)


def load_model(path: str) -> Model:
    """Loads a model from a pickle file."""

    with open(path, "rb") as handle:
        model = pickle.load(handle)

    return model



"""
# Random walk sampler
"""

from dataclasses import dataclass, field
from typing import ClassVar, Sequence

import jax
import jax.flatten_util

from .da import da_finalize, da_init, da_step
from .epoch import EpochState
from .kernel import (
    DefaultTransitionInfo,
    DefaultTuningInfo,
    ModelMixin,
    TransitionMixin,
    TransitionOutcome,
    TuningOutcome,
    WarmupOutcome,
)
from .mh import mh_step
from .pytree import register_dataclass_as_pytree
from .types import KeyArray, ModelState, Position, TuningInfo


@register_dataclass_as_pytree
@dataclass
class RWKernelState:
    """
    A dataclass for the state of a `RWKernel`, implementing the
    `liesel.goose.da.DAKernelState` protocol.
    """

    step_size: float
    error_sum: float = field(default=0.0, init=False)
    log_avg_step_size: float = field(default=0.0, init=False)
    mu: float = field(init=False)

    def __post_init__(self):
        da_init(self)


RWTransitionInfo = DefaultTransitionInfo
RWTuningInfo = DefaultTuningInfo


class RWKernel(ModelMixin, TransitionMixin[RWKernelState, RWTransitionInfo]):
    """
    A random walk kernel with Gaussian proposals, Metropolis-Hastings correction and
    dual averaging, implementing the `liesel.goose.types.Kernel` protocol.

    The kernel uses a default Metropolis-Hastings target acceptance probability
    of 0.234, which is optimal for a random walk sampler (in a certain sense). See
    Gelman et al., [Weak convergence and optimal scaling of random walk Metropolis
    algorithms (1997)](https://doi.org/10.1214/aoap/1034625254).
    """

    error_book: ClassVar[dict[int, str]] = {0: "no errors", 90: "nan acceptance prob"}
    needs_history: ClassVar[bool] = False
    identifier: str = ""

    def __init__(
        self,
        position_keys: Sequence[str],
        initial_step_size: float = 1.0,
        da_target_accept: float = 0.234,
        da_gamma: float = 0.05,
        da_kappa: float = 0.75,
        da_t0: int = 10,
    ):
        self._model = None
        self.position_keys = tuple(position_keys)
        self.initial_step_size = initial_step_size
        self.da_target_accept = da_target_accept
        self.da_gamma = da_gamma
        self.da_kappa = da_kappa
        self.da_t0 = da_t0

    def init_state(self, prng_key, model_state):
        """
        Initializes the kernel state.
        """

        return RWKernelState(step_size=self.initial_step_size)

    def _standard_transition(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[RWKernelState, DefaultTransitionInfo]:
        """
        Performs an MCMC transition *without* dual averaging.
        """

        key, subkey = jax.random.split(prng_key)
        step_size = kernel_state.step_size

        # random walk proposal
        position = self.position(model_state)
        flat_position, unravel_fn = jax.flatten_util.ravel_pytree(position)
        step = step_size * jax.random.normal(key, flat_position.shape)
        flat_proposal = flat_position + step
        proposal = unravel_fn(flat_proposal)

        # metropolis-hastings calibration
        info, model_state = mh_step(subkey, self.model, proposal, model_state)
        return TransitionOutcome(info, kernel_state, model_state)

    def _adaptive_transition(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[RWKernelState, DefaultTransitionInfo]:
        """
        Performs an MCMC transition *with* dual averaging.
        """

        outcome = self._standard_transition(prng_key, kernel_state, model_state, epoch)

        da_step(
            outcome.kernel_state,
            outcome.info.acceptance_prob,
            epoch.time_in_epoch,
            self.da_target_accept,
            self.da_gamma,
            self.da_kappa,
            self.da_t0,
        )

        return outcome

    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None = None,
    ) -> TuningOutcome[RWKernelState, DefaultTuningInfo]:
        """
        Currently does nothing.
        """

        info = RWTuningInfo(error_code=0, time=epoch.time)
        return TuningOutcome(info, kernel_state)

    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> RWKernelState:
        """
        Resets the state of the dual averaging algorithm.
        """

        da_init(kernel_state)
        return kernel_state

    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> RWKernelState:
        """
        Sets the step size as found by the dual averaging algorithm.
        """

        da_finalize(kernel_state)
        return kernel_state

    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        tuning_history: TuningInfo | None,
    ) -> WarmupOutcome[RWKernelState]:
        """
        Currently does nothing.
        """

        return WarmupOutcome(error_code=0, kernel_state=kernel_state)



# GENERATED FILE, DO NOT EDIT!

from tensorflow_probability.substrates.jax.distributions import *  # noqa: F401,F403

from .copulas import *  # noqa: F401,F403
from .smooth_prior import *  # noqa: F401,F403



"""
# A Rust-inspired Option type for Liesel and Goose
"""

from __future__ import annotations

import weakref
from typing import Callable, Generic, TypeVar

T = TypeVar("T")
U = TypeVar("U")


class Option(Generic[T]):
    """An Option type inspired by the Rust stdlib.

    See <https://doc.rust-lang.org/std/option/enum.Option.html> for descriptions
    of the methods.
    """

    def __init__(self, value: T | None) -> None:
        self._value = value

    @property
    def value(self) -> T | None:
        return self._value

    @value.setter
    def value(self, value: T | None) -> None:
        self._value = value

    @staticmethod
    def some(value: T) -> Option[T]:
        return Option(value)

    @staticmethod
    def none() -> Option[T]:
        return Option(None)

    def is_some(self) -> bool:
        return self.value is not None

    def is_none(self) -> bool:
        return self.value is None

    def expect(self, msg: str) -> T:
        if self.value is None:
            raise RuntimeError(msg)

        return self.value

    def unwrap(self) -> T:
        return self.expect(f"Trying to unwrap None from {repr(self)}")

    def unwrap_or(self, default: T) -> T:
        if self.value is None:
            return default

        return self.value

    def map(self, f: Callable[[T], U]) -> Option[U]:
        if self.value is None:
            return Option.none()

        return Option.some(f(self.value))

    def map_or(self, default: U, f: Callable[[T], U]) -> U:
        if self.value is None:
            return default

        return f(self.value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Option):
            return self.value == other.value
        else:
            raise NotImplementedError(
                f"Option cannot be compared to {type(other).__name__}"
            )

    def __repr__(self) -> str:
        return f"Option({repr(self.value)})"

    def __str__(self) -> str:
        return f"Option({self.value})"


class WeakOption(Option[T]):
    def __init__(self, value: T | None) -> None:
        self._ref = Option(value).map(lambda x: weakref.ref(x))

    @property
    def value(self) -> T | None:
        return self._ref.map_or(None, lambda x: x())

    @value.setter
    def value(self, value: T | None) -> None:
        self._ref = Option(value).map(lambda x: weakref.ref(x))

    def __repr__(self) -> str:
        return f"WeakOption({repr(self.value)})"

    def __str__(self) -> str:
        return f"WeakOption({self.value})"



"""
# Model interfaces
"""

import copy
from typing import Callable, Sequence

from .types import ModelState, Position

LogProbFunction = Callable[[ModelState], float]


class DictModel:
    """
    A model interface for a model state represented by a `dict[str, Array]`
    and a corresponding log-probability function.
    """

    def __init__(self, log_prob_fn: LogProbFunction):
        self._log_prob_fn = log_prob_fn

    def extract_position(
        self, position_keys: Sequence[str], model_state: ModelState
    ) -> Position:
        return Position({key: model_state[key] for key in position_keys})

    def update_state(self, position: Position, model_state: ModelState) -> ModelState:
        return model_state | position

    def log_prob(self, model_state: ModelState) -> float:
        return self._log_prob_fn(model_state)


class DataClassModel:
    """A model interface for a model state represented by a `dataclass`
    and a corresponding log-probability function."""

    def __init__(self, log_prob_fn: LogProbFunction):
        self._log_prob_fn = log_prob_fn

    def extract_position(
        self, position_keys: Sequence[str], model_state: ModelState
    ) -> Position:
        return Position({key: getattr(model_state, key) for key in position_keys})

    def log_prob(self, model_state: ModelState) -> float:
        return self._log_prob_fn(model_state)

    def update_state(self, position: Position, model_state: ModelState) -> ModelState:
        new_state = copy.copy(model_state)  # don't change the input
        for key, value in position.items():
            if hasattr(new_state, key):
                setattr(new_state, key, value)
            else:
                raise RuntimeError(
                    f"ModelState {model_state!r} does not have field with name {key}"
                )
        return new_state



"""
# Model visualization
"""

import matplotlib.pyplot as plt
import networkx as nx
from matplotlib.lines import Line2D


def plot_model(model, show=True, save_path=None, width=20, height=15, prog="dot"):
    """
    Plots a Liesel model.

    ## Parameters

    - `model`:
      The model to be plotted.

    - `show`:
      Whether to show the plot in a new window.

    - `save_path`:
      Path to save the plot.

      If not provided, the plot will not be saved.

    - `width`:
      Width of the plot in inches.

    - `height`:
      Height of the plot in inches.

    - `prog`:
      Layout parameter.

      Available layouts: circo, dot (the default), fdp, neato, osage, patchwork, sfdp,
      twopi.
    """

    _, axis, pos = _prepare_figure(model.graph, width, height, prog)
    _add_nodes_with_distribution_to_plot(model.graph, axis, pos)
    _add_nodes_without_distribution_to_plot(model.graph, axis, pos)
    _add_labels(model.graph, axis, pos)
    _draw_edges(model.graph, axis, pos)
    _add_legend(axis)

    if save_path:
        plt.savefig(save_path)
    if show:
        plt.show()


def _prepare_figure(graph, width, height, prog):
    """Prepares the figure for plotting."""

    fig, axis = plt.subplots()
    fig.set_size_inches(width, height)

    if _is_pygraphviz_installed():
        pos = nx.drawing.nx_agraph.graphviz_layout(graph, prog=prog)
    else:
        pos = nx.fruchterman_reingold_layout(graph)

    return fig, axis, pos


def _is_pygraphviz_installed():
    """Checks if pygraphviz is installed."""

    try:
        import pygraphviz  # noqa: F401
    except ImportError:
        return False
    else:
        return True


def _add_nodes_with_distribution_to_plot(graph, axis, pos):
    """Adds nodes with distribution to the figure."""

    nodes_with_distribution = {
        node: "#ffb4b4" if node.weak else "#b4b4ff"
        for node in graph.nodes
        if node.has_distribution
    }

    nx.draw_networkx_nodes(
        graph,
        pos,
        node_size=1200,
        node_color=nodes_with_distribution.values(),
        nodelist=nodes_with_distribution,
        node_shape="*",
        ax=axis,
    )


def _add_nodes_without_distribution_to_plot(graph, axis, pos):
    """Adds nodes without distribution to the figure."""

    nodes_without_distribution = {
        node: "#ffb4b4" if node.weak else "#b4b4ff"
        for node in graph.nodes
        if not node.has_distribution
    }

    nx.draw_networkx_nodes(
        graph,
        pos,
        node_size=500,
        node_color=nodes_without_distribution.values(),
        nodelist=nodes_without_distribution,
        node_shape="o",
        ax=axis,
    )


def _add_labels(graph, axis, pos):
    """Adds labels to the figure."""

    labels = {
        node: f"{type(node).__name__}\n{node.name}"
        if node.name is not None
        else node.role.name
        for node in pos
    }

    nx.draw_networkx_labels(graph, pos, labels=labels, ax=axis, font_size=10)


def _draw_edges(graph, axis, pos):
    """Adds edges to the figure."""

    nx.draw_networkx_edges(
        graph,
        pos,
        edgelist=graph.edges,
        edge_color="#aaaaaa",
        arrows=True,
        ax=axis,
        node_size=500,
    )


def _add_legend(axis):
    """Adds a legend to the figure."""

    legend_elements = [
        Line2D([0], [0], color="#b4b4ff", lw=4, label="Strong"),
        Line2D([0], [0], color="#ffb4b4", lw=4, label="Weak"),
        Line2D(
            [0],
            [0],
            marker="*",
            color="w",
            label="With distribution",
            markerfacecolor="k",
            markersize=18,
        ),
        Line2D(
            [0],
            [0],
            marker="o",
            color="w",
            label="Without distribution",
            markerfacecolor="k",
            markersize=12,
        ),
    ]

    axis.legend(handles=legend_elements, loc="best")



"""
# Standard warmup schemes
"""

from functools import partial

from .epoch import EpochConfig, EpochType

_EpochConfig = partial(EpochConfig, thinning=1, optional=None)


def stan_epochs(
    warmup_duration: int = 1000,
    posterior_duration: int = 1000,
    init_duration: int = 75,
    term_duration: int = 50,
    base_duration: int = 25,
) -> list[EpochConfig]:
    """
    Sets up a list of `liesel.goose.epoch.EpochConfig`'s,
    following the Stan Development Team, [Stan Reference Manual (2021), Chapter 15.2](
    https://mc-stan.org/docs/2_28/reference-manual/hmc-algorithm-parameters.html).

    ## Parameters

    - `warmup_duration`: The number of warmup samples.
    - `posterior_duration`: The number of posterior samples.
    - `init_duration`: The number of samples in the *initial fast* adaptation epoch.
    - `term_duration`: The number of samples in the *final fast* adaptation epoch.
    - `base_duration`: The number of samples in the *first slow* adaptation epoch.
    """

    if warmup_duration < 20:
        raise ValueError("warmup_duration too short (< 20)")

    if warmup_duration < init_duration + term_duration + base_duration:
        raise ValueError(
            "warmup_duration too short "
            "(< init_duration + term_duration + base_duration)"
        )

    epochs = [_EpochConfig(EpochType.INITIAL_VALUES, duration=1)]
    epochs.append(_EpochConfig(EpochType.FAST_ADAPTATION, init_duration))

    time_left = warmup_duration - init_duration - term_duration
    this_time = base_duration

    while 3 * this_time <= time_left:
        epochs.append(_EpochConfig(EpochType.SLOW_ADAPTATION, this_time))
        time_left -= this_time
        this_time *= 2

    epochs.append(_EpochConfig(EpochType.SLOW_ADAPTATION, time_left))
    epochs.append(_EpochConfig(EpochType.FAST_ADAPTATION, term_duration))
    epochs.append(_EpochConfig(EpochType.POSTERIOR, posterior_duration))

    return epochs



"""
# Inverse mass matrix tuner

This module uses the error codes 70-79.
"""

import jax
import jax.numpy as jnp

from .types import Array, Position

_vravel = jax.vmap(jnp.ravel, in_axes=0, out_axes=0)


def _history_to_matrix(history: Position) -> Array:
    return jnp.column_stack([_vravel(x) for x in history.values()])


def tune_inv_mm_diag(history: Position) -> Array:
    """
    Tunes an inverse mass vector with the sample variances of the history.
    """

    matrix = _history_to_matrix(history)
    var = jnp.var(matrix, axis=0, ddof=1)
    var = jnp.atleast_1d(var)
    var = var + 0.001

    return var


def tune_inv_mm_full(history: Position) -> Array:
    """
    Tunes an inverse mass matrix with the sample variance-covariance matrix
    of the history.
    """

    matrix = _history_to_matrix(history)
    cov = jnp.cov(matrix, rowvar=False)
    cov = jnp.atleast_2d(cov)

    # stan regularization, see:
    # https://github.com/stan-dev/stan/blob/v2.28.2/src/stan/mcmc/covar_adaptation.hpp#L27-L29

    # n = matrix.shape[0]
    # cov = (n / (n + 5.0)) * cov
    # cov = cov.at[jnp.diag_indices_from(cov)].add(0.001 * (5.0 / (n + 5.0)))
    cov = cov.at[jnp.diag_indices_from(cov)].add(0.001)

    return cov



"""
# MCMC epochs
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Iterable, cast

from .pytree import register_dataclass_as_pytree
from .types import PyTree


class EpochType(IntEnum):
    """Indicates which MCMC phase the epoch is part of."""

    INITIAL_VALUES = 0
    FAST_ADAPTATION = 1
    SLOW_ADAPTATION = 2
    BURNIN = 3
    POSTERIOR = 4

    @staticmethod
    def is_adaptation(epoch_type: EpochType) -> bool:
        """
        Returns `True` if the epoch is part of the adaptation phase.
        Implemented as a static method to make it jittable.
        """
        lhs = EpochType.INITIAL_VALUES < epoch_type
        rhs = epoch_type < EpochType.BURNIN
        # using `*` instead of `and` for jax.jit
        # `cast` is a no-op used to satisfy mypy
        return cast(bool, lhs * rhs)

    @staticmethod
    def is_warmup(epoch_type: EpochType) -> bool:
        """
        Returns `True` if the epoch is part of the warmup phase.
        Implemented as a static method to make it jittable.
        """
        rhs = EpochType.INITIAL_VALUES < epoch_type
        lhs = epoch_type < EpochType.POSTERIOR
        # using `*` instead of `and` for jax.jit
        # `cast` is a no-op used to satisfy mypy
        return cast(bool, lhs * rhs)


@register_dataclass_as_pytree
@dataclass
class EpochConfig:
    type: EpochType
    duration: int
    thinning: int
    optional: PyTree | None

    def to_state(self, nth_epoch: int, time_before_epoch: int) -> EpochState:
        return EpochState(
            config=self,
            nth_epoch=nth_epoch,
            time=time_before_epoch,
            time_before_epoch=time_before_epoch,
            time_in_epoch=0,
        )


@register_dataclass_as_pytree
@dataclass
class EpochState:
    config: EpochConfig
    nth_epoch: int
    time: int
    time_before_epoch: int
    time_in_epoch: int

    def time_left(self) -> int:
        return self.config.duration - self.time_in_epoch

    def advance_time(self, by: int):
        self.time = self.time + by
        self.time_in_epoch = self.time_in_epoch + by


class EpochManager:
    """
    Manages `EpochConfig` objects.

    A sequence of `EpochConfig` objects can be handed to the manager either
    during initialization or later with the `append` method. The manager creates
    a new `EpochState` object with properly initialized time values.

    Furthermore, the `EpochManager` enforces this invariant:

    - An epoch must be of duration at least 1.
    - An initial values epoch must be of duration exactly 1.
    - The first epoch must be an initial values epoch.
    - No other than the first epoch may be an initial values epoch.
    - A posterior epoch may not be followed by a non-posterior epoch.
    - Temporary: Thinning must be 1.

    The limitation that thinning must be 1 will probably be removed in the
    future. To do so, we need to figure out how kernels relying on history
    tuning can be tuned if the history is thinned.
    """

    def __init__(self, configs: Iterable[EpochConfig] | None):
        self._configs: list[EpochConfig] = []
        self._next_epoch_ptr = 0
        self._next_start_time: int = 0
        self._nth_epoch: int = 0

        if configs is not None:
            for config in configs:
                self.append(config)

    def append(self, config: EpochConfig):
        """Appends an `EpochConfig` to the list of epochs."""
        if not self._configs and config.type != EpochType.INITIAL_VALUES:
            raise RuntimeError("First epoch must be of type INITIAL_VALUES")

        if config.type == EpochType.INITIAL_VALUES:
            if self._configs:
                raise RuntimeError("Only the first epoch may be of type INITIAL_VALUES")

            if config.duration != 1:
                raise RuntimeError("Epochs of type INITIAL_VALUES must have duration 1")

        if (
            EpochType.is_warmup(config.type)
            and self._configs[-1].type == EpochType.POSTERIOR
        ):
            raise RuntimeError(
                "Warmup epochs may not follow an epoch of type POSTERIOR"
            )

        if config.duration < 1:
            raise RuntimeError("Duration must be greater than or equal to 1")

        if config.thinning != 1:
            raise RuntimeError(
                "Thinning is currently not supported and must be set to 1"
            )

        self._configs.append(config)

    def has_more(self) -> bool:
        """
        Returns true if there are epoch configs that have not been returned yet
        by the `next` method, and false otherwise.
        """
        return self._next_epoch_ptr < len(self._configs)

    def next(self) -> EpochState:
        """
        Returns the next epoch with an initialized state.

        Raises a `RuntimeError` if there are no more epoch configs to return.
        """
        if self.has_more():
            config = self._configs[self._next_epoch_ptr]
            start_time = self._next_start_time
            self._next_start_time += config.duration
            state = config.to_state(self._next_epoch_ptr, start_time)
            self._next_epoch_ptr += 1
            return state
        else:
            raise RuntimeError("No epochs in manager")



"""
# Diagnostic plots of the posterior samples
"""

from collections.abc import Sequence
from typing import Any

import arviz
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from liesel.goose.engine import SamplingResult
from liesel.goose.summary_m import (
    adjust_dimensions,
    move_col_first,
    raise_dimension_error,
    validate_chain_indices,
    validate_param_indices,
    validate_params,
)


def subparam_chains_to_df(
    subparam_chains: np.ndarray, param_index: int
) -> pd.DataFrame:
    """
    Convert array of posterior samples for a single subparameter (e.g. beta_0) to a
    pandas data frame.
    """

    subparam_df = (
        pd.DataFrame(subparam_chains)
        .melt(ignore_index=False)
        .rename(columns={"variable": "iteration"})
        .reset_index()
        .rename(columns={"index": "chain_index"})
        .sort_values(by=["chain_index", "iteration"], ignore_index=True)
        .assign(param_index=param_index)
    )

    return move_col_first(subparam_df, colname="param_index")


def preprocess_param_chains(
    posterior_samples: dict[str, jnp.DeviceArray], param: str
) -> np.ndarray:
    """Convert array of posteror samples for each parameter to equal dimensions."""

    param_chains = np.array(posterior_samples[param])
    num_dim = param_chains.ndim

    raise_dimension_error(param, num_dim)
    param_chains = adjust_dimensions(param_chains, num_dim)
    return param_chains


def convert_to_sequence(indices: int | Sequence[int]) -> Sequence[int]:
    """Convert integer parameter and chain indices to list or tuple."""

    if isinstance(indices, int):
        indices = [indices]

    return indices


def merge_cols(df: pd.DataFrame, col1: str, col2: str) -> pd.DataFrame:
    """
    Combine param and param_index column to single column. Transforms e.g. beta and 0 to
    beta_0.
    """

    df[col1] = df[col1].astype(str) + "_" + df[col2].astype(str)
    return df


def filter_param_df(
    param_df: pd.DataFrame,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
) -> pd.DataFrame:
    """
    Filters the plotting data to contain only the specified parameter and chain indices.
    Output data contains not more than `max_chains` chain indices.
    """

    if chain_indices is not None:
        chain_indices = convert_to_sequence(chain_indices)
        param_df = param_df.loc[param_df["chain_index"].isin(chain_indices)]

    if param_indices is not None:
        param_indices = convert_to_sequence(param_indices)
        param_df = param_df.loc[param_df["param_index"].isin(param_indices)]

    if max_chains is not None and param_df["chain_index"].nunique() > max_chains:
        last_chain_index = sorted(param_df["chain_index"].unique())[max_chains - 1]
        param_df = param_df.loc[param_df["chain_index"] <= last_chain_index]

    return param_df


def postprocess_param_df(
    param_df: pd.DataFrame,
    param: str,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
) -> pd.DataFrame:
    """
    Combines parameter and parameter index column and filters chain and parameter
    indices.
    """

    num_original_chains = param_df["chain_index"].nunique()
    num_original_subparams = param_df["param_index"].nunique()

    if num_original_subparams > 1:
        param_df = merge_cols(param_df, col1="param", col2="param_index")

    chain_indices = validate_chain_indices(chain_indices, num_original_chains)
    param_indices = validate_param_indices(param_indices, num_original_subparams, param)
    param_df = filter_param_df(param_df, param_indices, chain_indices, max_chains)

    return param_df.drop(columns="param_index")


def collect_subparam_dfs(
    posterior_samples: dict[str, jnp.DeviceArray],
    param: str,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
) -> pd.DataFrame:
    """
    Combines individual data frames for each subparameter into a single data frame for
    each parameter.
    """

    param_chains = preprocess_param_chains(posterior_samples, param)

    param_df = (
        pd.concat(
            [
                subparam_chains_to_df(param_chains[..., param_index], param_index)
                for param_index in range(param_chains.shape[-1])
            ]
        )
        .assign(param=param)
        .reset_index(drop=True)
    )

    param_df = move_col_first(param_df, colname="param")

    return postprocess_param_df(
        param_df, param, param_indices, chain_indices, max_chains
    )


def collect_param_dfs(
    results: SamplingResult,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    include_warmup: bool = False,
) -> pd.DataFrame:
    """Combines individual data frames for each parameter into a single data frame."""

    if include_warmup:
        samples = results.get_samples()
    else:
        samples = results.get_posterior_samples()
    params = validate_params(samples, params)

    return pd.concat(
        [
            collect_subparam_dfs(
                samples, param, param_indices, chain_indices, max_chains
            )
            for param in params
        ]
    ).reset_index(drop=True)


def setup_plot_df(
    results: SamplingResult,
    params: str | list[str] | None,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
    include_warmup: bool = False,
) -> pd.DataFrame:
    """Provides data input for all plotting functions."""

    return collect_param_dfs(
        results,
        params,
        param_indices,
        chain_indices,
        max_chains,
        include_warmup,
    ).astype({"chain_index": "category"})


def set_plot_cols(plot_df: pd.DataFrame, ncol: int) -> int:
    """Determines number of facets within each row of the grid."""

    num_subparams = plot_df["param"].nunique()
    return min(ncol, num_subparams)


def set_aesthetics(
    g: sns.FacetGrid, title: str | None, title_spacing: float, xlabel: str, ylabel: str
) -> sns.FacetGrid:
    """Adds titles, labels and correct spacing between facets."""

    g.set_titles(col_template="{col_name}")
    g.set_axis_labels(x_var=xlabel, y_var=ylabel)
    g.tight_layout()

    g.legend.set_title("Chain")

    if title is not None:
        g.fig.suptitle(title)
        g.fig.subplots_adjust(top=title_spacing)

    return g


def save_figure(g: sns.FacetGrid | None = None, save_path: str | None = None) -> None:
    """Saves plot to file."""

    if save_path is not None:
        if g is not None:
            g.fig.savefig(save_path)
        else:
            plt.savefig(save_path)
    else:
        plt.show()


def plot_trace(
    results: SamplingResult,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    title: str | None = None,
    title_spacing: float = 0.85,
    xlabel: str = "Iteration",
    style: str = "whitegrid",
    color_palette: str | list[str] | dict[int, str] | None = None,
    ncol: int = 3,
    height: int = 3,
    aspect_ratio: int = 1,
    save_path: str | None = None,
    include_warmup: bool = False,
    **kwargs,
) -> sns.FacetGrid:
    """
    Visualize posterior samples over time with a trace plot.

    ## Parameters

    - `results`:
      Result object of the sampling process. Must have a method
      `get_posterior_samples()` which extracts all samples from the posterior
      distribution.

    - `params`:
      Names of the model parameters that are contained in the plot. Must coincide with
      the dictionary keys of the `Position` with the posterior samples. If `None`,
      all parameters are included.

    - `param_indices`:
      Indices of each model parameter that are contained in the plot. Selects e.g.
      `beta_0` out of a `beta` parameter vector.A single index can be specified as an
      integer or a sequence containing one integer. If `None`, all subparameters are
      included.

    - `chain_indices`:
      Indices of chains for each model subparameter that are contained in the plot.
      Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
      specified as an integer or a sequence containing one integer. If `None`, all
      chains are included.

    - `max_chains`:
      Upper bound how many chains are included within each subplot/facet. Avoids
      overplotting. If `None`, all chains contained in the `results` input are plotted.
      Always starts chain selection from the lowest chain index upwards. For selecting
      specific chains use the argument `chain_indices`.

    - `title`:
      Plot title.

    - `title_spacing`:
      Determines the margin/whitespace between the plot title (set with
      `fig.suptitle()`) and the first row of subplots/facets. Passed to the `top`
      argument of `fig.subplots_adjust()`.

    - `xlabel`:
      Label of the x-axis.

    - `style`:
      Passed to the `style` argument of `sns.set_theme()`. Valid options are `darkgrid`,
      `whitegrid`, `dark`, `white`, and `ticks`.

    - `color_palette`:
      Passed to the palette argument of `sns.relplot()`. String values must be valid
      inputs of `sns.color_palette()` such as a seaborn color palette or a matplotlib
      colormap. Custom colors can be set with a list of color strings or a dictionary
      with the chain indices as keys and color strings as values. The number of color
      strings must coincide with the number of plotted chains. If `None`, the default
      `tab10` matplotlib colormap is chosen.

    - `ncol`:
      Number of subplots/facets within each row of the grid.

    - `height`:
      Height in inches of each subplot/facet within the grid.

    - `aspect_ratio`:
      Ratio of width / height of each subplot/facet within the grid,
      i.e. `width = aspect_ratio * height`.

    - `save_path`:
      File path where the plot is saved.

    - `include_warmup:
       Include the warmup samples in the trace plot.

    - `**kwargs`:
      Further keyword arguments passed to the seaborn `relplot()` function.

    ## Returns

    A seaborn `FacetGrid`.
    """

    # NOTE: Docstring duplications
    # The entries `results` to `max_chains` are shared with the `summary()`
    # and all user plotting functions.
    # The entries `title` to `save_path` are shared with `plot_density()`, `plot_cor()`
    # and partially with `plot_param()`.

    sns.set_theme(style=style)
    plot_df = setup_plot_df(
        results, params, param_indices, chain_indices, max_chains, include_warmup
    )

    g = sns.relplot(
        data=plot_df,
        kind="line",
        x="iteration",
        y="value",
        hue="chain_index",
        col="param",
        col_wrap=set_plot_cols(plot_df, ncol),
        facet_kws=dict(sharex=True, sharey=False),
        palette=color_palette,
        height=height,
        aspect=aspect_ratio,
        **kwargs,
    )

    g = set_aesthetics(g, title, title_spacing, xlabel, ylabel="")

    save_figure(g, save_path)
    return g


def plot_density(
    results: SamplingResult,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    title: str | None = None,
    title_spacing: float = 0.85,
    xlabel: str = "Value",
    style: str = "whitegrid",
    color_palette: str | list[str] | dict[int, str] | None = None,
    ncol: int = 3,
    height: int = 3,
    aspect_ratio: int = 1,
    save_path: str | None = None,
    **kwargs,
) -> sns.FacetGrid:
    """
    Visualize posterior distributions with a density plot.

    ## Parameters

    - `results`:
      Result object of the sampling process. Must have a method
      `get_posterior_samples()` which extracts all samples from the posterior
      distribution.

    - `params`:
      Names of the model parameters that are contained in the plot. Must coincide with
      the dictionary keys of the `Position` with the posterior samples. If `None`,
      all parameters are included.

    - `param_indices`:
      Indices of each model parameter that are contained in the plot. Selects e.g.
      `beta_0` out of a `beta` parameter vector.A single index can be specified as an
      integer or a sequence containing one integer. If `None`, all subparameters are
      included.

    - `chain_indices`:
      Indices of chains for each model subparameter that are contained in the plot.
      Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
      specified as an integer or a sequence containing one integer. If `None`, all
      chains are included.

    - `max_chains`:
      Upper bound how many chains are included within each subplot/facet. Avoids
      overplotting. If `None`, all chains contained in the `results` input are plotted.
      Always starts chain selection from the lowest chain index upwards. For selecting
      specific chains use the argument `chain_indices`.

    - `title`:
      Plot title.

    - `title_spacing`:
      Determines the margin/whitespace between the plot title (set with
      `fig.suptitle()`) and the first row of subplots/facets. Passed to the `top`
      argument of `fig.subplots_adjust()`.

    - `xlabel`:
      Label of the x-axis.

    - `style`:
      Passed to the `style` argument of `sns.set_theme()`. Valid options are `darkgrid`,
      `whitegrid`, `dark`, `white`, and `ticks`.

    - `color_palette`:
      Passed to the palette argument of `sns.displot()`. String values must be valid
      inputs of `sns.color_palette()` such as a seaborn color palette or a matplotlib
      colormap. Custom colors can be set with a list of color strings or a dictionary
      with the chain indices as keys and color strings as values. The number of color
      strings must coincide with the number of plotted chains. If `None`, the default
      `tab10` matplotlib colormap is chosen.

    - `ncol`:
      Number of subplots/facets within each row of the grid.

    - `height`:
      Height in inches of each subplot/facet within the grid.

    - `aspect_ratio`:
      Ratio of width / height of each subplot/facet within the grid,
      i.e. `width = aspect_ratio * height`.

    - `save_path`:
      File path where the plot is saved.

    - `**kwargs`:
      Further keyword arguments passed to the seaborn `displot()` function.

    ## Returns

    A seaborn `FacetGrid`.
    """

    # NOTE: Docstring duplications
    # The entries `results` to `max_chains` are shared with the `summary()`
    # and all user plotting functions.
    # The entries `title` to `save_path` are shared with `plot_trace()`, `plot_cor()`
    # and partially with `plot_param()`.

    sns.set_theme(style=style)
    plot_df = setup_plot_df(results, params, param_indices, chain_indices, max_chains)

    g = sns.displot(
        data=plot_df,
        kind="kde",
        x="value",
        y=None,
        hue="chain_index",
        col="param",
        col_wrap=set_plot_cols(plot_df, ncol),
        facet_kws=dict(sharex=False, sharey=False),
        palette=color_palette,
        height=height,
        aspect=aspect_ratio,
        **kwargs,
    )

    g = set_aesthetics(g, title, title_spacing, xlabel, ylabel="")

    save_figure(g, save_path)
    return g


def compute_max_lags(
    plot_df: pd.DataFrame,
    max_lags: int | None,
) -> int:
    """
    Determines number time lags that are shown on the x-axis of the autocorrelation
    plot.
    """

    num_iterations = plot_df["iteration"].max().item()
    max_lags = np.min([num_iterations, 30]) if max_lags is None else max_lags
    return max_lags


def plot_cor(
    results: SamplingResult,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    max_lags: int | None = None,
    title: str | None = None,
    title_spacing: float = 0.85,
    xlabel: str = "Lag",
    style: str = "whitegrid",
    color_palette: str | list[str] | dict[int, str] | None = None,
    ncol: int = 3,
    height: int = 3,
    aspect_ratio: int = 1,
    save_path: str | None = None,
    **kwargs,
) -> sns.FacetGrid:
    """
    Visualize autocorrelations of posterior samples.

    ## Parameters

    - `results`:
      Result object of the sampling process. Must have a method
      `get_posterior_samples()` which extracts all samples from the posterior
      distribution.

    - `params`:
      Names of the model parameters that are contained in the plot. Must coincide with
      the dictionary keys of the `Position` with the posterior samples. If `None`,
      all parameters are included.

    - `param_indices`:
      Indices of each model parameter that are contained in the plot. Selects e.g.
      `beta_0` out of a `beta` parameter vector.A single index can be specified as an
      integer or a sequence containing one integer. If `None`, all subparameters are
      included.

    - `chain_indices`:
      Indices of chains for each model subparameter that are contained in the plot.
      Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
      specified as an integer or a sequence containing one integer. If `None`, all
      chains are included.

    - `max_chains`:
      Upper bound how many chains are included within each subplot/facet. Avoids
      overplotting. If `None`, all chains contained in the `results` input are plotted.
      Always starts chain selection from the lowest chain index upwards. For selecting
      specific chains use the argument `chain_indices`.

    - `max_lags`:
      Maximum number of time lags shown on the x-axis of the autocorrelation plot. If
      `None`, the minimum of the chain lengths and 30 is chosen.

    - `title`:
      Plot title.

    - `title_spacing`:
      Determines the margin/whitespace between the plot title (set with
      `fig.suptitle()`) and the first row of subplots/facets. Passed to the `top`
      argument of `fig.subplots_adjust()`.

    - `xlabel`:
      Label of the x-axis.

    - `style`:
      Passed to the `style` argument of `sns.set_theme()`. Valid options are `darkgrid`,
      `whitegrid`, `dark`, `white`, and `ticks`.

    - `color_palette`:
      Passed to the palette argument of `sns.FacetGrid()`. String values must be valid
      inputs of `sns.color_palette()` such as a seaborn color palette or a matplotlib
      colormap. Custom colors can be set with a list of color strings or a dictionary
      with the chain indices as keys and color strings as values. The number of color
      strings must coincide with the number of plotted chains. If `None`, the default
      `tab10` matplotlib colormap is chosen.

    - `ncol`:
      Number of subplots/facets within each row of the grid.

    - `height`:
      Height in inches of each subplot/facet within the grid.

    - `aspect_ratio`:
      Ratio of width / height of each subplot/facet within the grid,
      i.e. `width = aspect_ratio * height`.

    - `save_path`:
      File path where the plot is saved.

    - `**kwargs`:
      Further keyword arguments passed to the seaborn `FacetGrid()` function.

    ## Returns

    A seaborn `FacetGrid`.
    """

    # NOTE: Docstring duplications
    # The entries `results` to `max_chains` are shared with the `summary()`
    # and all user plotting functions.
    # The entries `title` to `save_path` are shared with `plot_trace()`,
    # `plot_density()` and partially with `plot_param()`.
    # The entry `max_lags` is shared with `plot_param()`.

    sns.set_theme(style=style)
    plot_df = setup_plot_df(results, params, param_indices, chain_indices, max_chains)
    max_lags = compute_max_lags(plot_df, max_lags)

    def do_acor_plot(x, maxlags, **kwargs):
        x = np.asarray(x)
        acor = arviz.autocorr(x)[..., 0:maxlags]
        return sns.lineplot(x=range(maxlags), y=acor, **kwargs)

    g = (
        sns.FacetGrid(
            data=plot_df,
            hue="chain_index",
            col="param",
            col_wrap=set_plot_cols(plot_df, ncol),
            palette=color_palette,
            height=height,
            aspect=aspect_ratio,
            **kwargs,
        )
        .map(
            do_acor_plot,
            "value",
            maxlags=max_lags,
        )
        .set(
            xlim=(0, max_lags),
            ylim=(-0.2, 1.1),
        )
        .add_legend()
    )

    g = set_aesthetics(g, title, title_spacing, xlabel, ylabel="Autocorrelation")

    save_figure(g, save_path)
    return g


def raise_multi_param_error(plot_df: pd.DataFrame, param: str) -> None:
    """
    `plot_param()` function can only display all three diagnostic plots for a single
    subparameter. Throws an informative error otherwise.
    """

    if plot_df["param"].nunique() > 1:
        raise ValueError(
            f"{param} has more than one index. "
            f"Please specify a single `param_index` for plotting."
        )


def set_colors(plot_df: pd.DataFrame, color_list: list[str] | None) -> list[str]:
    """Determines colors of different chains in each plot."""

    num_chains = plot_df["chain_index"].nunique()

    if color_list is None:
        # default matplotlib and seaborn colors with 10 elements
        color_list = sns.color_palette()

        # make default color list sufficiently long
        if num_chains > 10:
            color_list = color_list * (num_chains // 10 + 1)

        color_list = color_list[:num_chains]  # type: ignore

    return color_list


def setup_grid(
    figure_size: tuple[int | float, int | float]
) -> tuple[plt.Figure, Any, Any, Any]:
    """
    Initializes plotting grid with one large subplot for the trace plot and two smaller
    subplots for the density and autocorrelation plot.
    """

    fig = plt.figure(figsize=figure_size)
    ax1 = plt.subplot2grid(shape=(2, 2), loc=(0, 0), colspan=2)
    ax2 = plt.subplot2grid(shape=(2, 2), loc=(1, 0))
    ax3 = plt.subplot2grid(shape=(2, 2), loc=(1, 1))
    return fig, ax1, ax2, ax3


def add_lineplot(plot_df: pd.DataFrame, ax: Any, color_list: list[str]) -> None:
    """Adds trace plot to plotting grid."""

    sns.lineplot(
        data=plot_df,
        x="iteration",
        y="value",
        hue="chain_index",
        palette=color_list,
        ax=ax,
        legend="full",
    ).set(xlabel="Iteration", ylabel="")


def add_kdeplot(plot_df: pd.DataFrame, ax: Any, color_list: list[str]) -> None:
    """Adds density plot to plotting grid."""

    sns.kdeplot(
        data=plot_df,
        x="value",
        hue="chain_index",
        palette=color_list,
        ax=ax,
        legend=False,
    ).set(xlabel="Value", ylabel="")


def add_corplot(
    plot_df: pd.DataFrame, ax: Any, max_lags: int | None, color_list: list[str]
) -> None:
    """Adds correlation plot to plotting grid."""

    max_lags = compute_max_lags(plot_df, max_lags)

    for chain_index, col in zip(plot_df["chain_index"].unique(), color_list):
        x = np.asarray(plot_df.loc[plot_df["chain_index"] == chain_index]["value"])
        acor = arviz.autocorr(x)[0:max_lags]

        sns.lineplot(
            x=range(max_lags),
            y=acor,
            marker="",
            linestyle="-",
            color=col,
            ax=ax,
        )

        ax.set(xlim=(0, max_lags), ylim=(-0.2, 1.1), xlabel="Lag", ylabel="")


def get_title(plot_df: pd.DataFrame, title: str | None) -> str:
    """Sets either a custom or the default plot title."""

    default_title = f"Diagnostic plots for '{plot_df['param'][0]}'"
    return title if title is not None else default_title


def plot_param(
    results: SamplingResult,
    param: str,
    param_index: int | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    max_lags: int | None = None,
    title: str | None = None,
    title_spacing: float = 0.9,
    style: str = "whitegrid",
    color_list: list[str] | None = None,
    figure_size: tuple[int | float, int | float] = (9, 6),
    # default values chosen for default figure size of (9, 6)
    legend_position: tuple[float, float] = (1.2, 0.4),
    save_path: str | None = None,
) -> None:
    """
    Visualize trace plot, density plot and autocorrelation plot of a single
    subparameter.

    ## Parameters

    - `results`:
      Result object of the sampling process. Must have a method
      `get_posterior_samples()` which extracts all samples from the posterior
      distribution.

    - `param`:
      Name of a single model parameter that is contained in the plot. Must coincide with
      one dictionary key of the `Position` with the posterior samples.

    - `param_index`:
      A single index of the selected model parameter that is contained in the plot.
      Selects e.g. `beta_0` out of a `beta` parameter vector. Can be specified as an
      integer or as a sequence containing one integer. If `None`, the parameter is
      assumed to have only a single index.

    - `chain_indices`:
      Indices of chains for each model subparameter that are contained in the plot.
      Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
      specified as an integer or a sequence containing one integer. If `None`, all
      chains are included.

    - `max_chains`:
      Upper bound how many chains are included within each subplot/facet. Avoids
      overplotting. If `None`, all chains contained in the `results` input are plotted.
      Always starts chain selection from the lowest chain index upwards. For selecting
      specific chains use the argument `chain_indices`.

    - `max_lags`:
      Maximum number of time lags shown on the x-axis of the autocorrelation plot. If
      `None`, the minimum of the chain lengths and 30 is chosen.

    - `title`:
      Plot title.

    - `title_spacing`:
      Determines the margin/whitespace between the plot title (set with
      `fig.suptitle()`) and the first row of subplots/facets. Passed to the `top`
      argument of `fig.subplots_adjust()`.

    - `style`:
      Passed to the `style` argument of `sns.set_theme()`. Valid options are `darkgrid`,
      `whitegrid`, `dark`, `white`, and `ticks`.

    - `color_list`:
      Determines the chain colors for all three subplots. Custom colors can be passed
      with a list of color strings. The length of the list must match the number of
      chains. If `None`, the default `tab10` matplotlib colormap is chosen.

    - `figure_size`:
      Size of the entire plot grid. Passed to the `figsize` argument of `plt.figure()`.
      When changing the figure size consider changing the `legend_position` as well.
      Generally, a ratio of 3:2 is recommended.

    - `legend_position`:
      Determines the color legend position. Coordinates are relative to the upper panel
      within the plot grid. The first coordinate specifies the horizontal, the second
      coordinate the vertical position. Might require an adjustment when changing the
      `figure_size` values or the number of chains.

    - `save_path`:
      File path where the plot is saved.
    """

    # NOTE: Docstring duplications
    # The entries `results`, `chain_indices` and `max_chains` are shared with the
    # `summary()` and all user plotting functions.
    # The entries `title`, `title_spacing`, `style` and `save_path` are shared with
    # `plot_trace()`, `plot_density()` and `plot_cor()`.
    # The entry `max_lags` is shared with `plot_cor()`.

    sns.set_theme(style=style)
    plot_df = setup_plot_df(results, param, param_index, chain_indices, max_chains)
    raise_multi_param_error(plot_df, param)
    color_list = set_colors(plot_df, color_list)

    fig, ax1, ax2, ax3 = setup_grid(figure_size)
    add_lineplot(plot_df, ax1, color_list)
    add_kdeplot(plot_df, ax2, color_list)
    add_corplot(plot_df, ax3, max_lags, color_list)
    ax1.legend(title="Chain", bbox_to_anchor=legend_position, frameon=False)

    fig.tight_layout()
    fig.suptitle(get_title(plot_df, title))
    fig.subplots_adjust(top=title_spacing)

    save_figure(save_path=save_path)



"""
# Dual averaging

This module uses the error codes 80-89.
"""

from typing import Protocol

import jax.numpy as jnp


class DAKernelState(Protocol):
    """
    A protocol for a kernel state with dual averaging support. For an introduction to
    dual averaging, see:

    - Colin Carroll, [Step Size Adaptation in Hamiltonian Monte Carlo (2019)](
      https://colindcarroll.com/2019/04/21/step-size-adaptation-in-hamiltonian-monte-carlo).
    - Stan Development Team, [Stan Reference Manual (2021), Chapter 15.2](
      https://mc-stan.org/docs/2_28/reference-manual/hmc-algorithm-parameters.html).
    """

    step_size: float
    """The step size of the kernel."""

    error_sum: float
    """The error sum of the acceptance probability. Should not be set by the user,
    but is used by the `da_step` function."""

    log_avg_step_size: float
    """The logarithm of the average step size. Should not be set by the user, but is
    used by the `da_step` function."""

    mu: float
    """The bias of the step size proposals. Should not be set by the user, but is
    used by the `da_step` function."""


def da_init(kernel_state: DAKernelState) -> None:
    """
    Initializes (or resets) a `DAKernelState`. Returns `None` and should be called for
    the side effect on the `kernel_state` argument.
    """

    kernel_state.error_sum = 0.0
    kernel_state.log_avg_step_size = jnp.log(kernel_state.step_size)
    kernel_state.mu = jnp.log(10.0 * kernel_state.step_size)


def da_step(
    kernel_state: DAKernelState,
    acceptance_prob: float,
    time_in_epoch: int,
    target_accept: float = 0.8,
    gamma: float = 0.05,
    kappa: float = 0.75,
    t0: int = 10,
) -> None:
    """
    Performs an dual averaging update on a `DAKernelState`. Returns `None` and should
    be called for the side effect on the `kernel_state` argument.

    ## Parameters

    - `kernel_state`: A kernel state implementing the `DAKernelState` protocol.
    - `acceptance_prob`: The acceptance probability of this MCMC iteration.
    - `time_in_epoch`: The number of completed MCMC iterations in this epoch.
    - `target_accept`: The target acceptance probability.
    - `gamma`: The adaptation regularization scale.
    - `kappa`: The adaptation relaxation exponent.
    - `t0`: The adaptation iteration offset.
    """

    ks = kernel_state
    t = time_in_epoch + 1
    eta = t ** (-kappa)

    ks.error_sum += target_accept - acceptance_prob
    log_step_size = ks.mu - (ks.error_sum * jnp.sqrt(t)) / (gamma * (t0 + t))
    ks.step_size = jnp.exp(log_step_size)

    ks.log_avg_step_size = (1 - eta) * ks.log_avg_step_size + eta * log_step_size


def da_finalize(kernel_state: DAKernelState) -> None:
    """
    Sets the new step size in a `DAKernelState`. Returns `None` and should be called
    for the side effect on the `kernel_state` argument.
    """

    kernel_state.step_size = jnp.exp(kernel_state.log_avg_step_size)



"""
# Iteratively weighted least squares (IWLS) sampler
"""

from dataclasses import dataclass, field
from typing import Callable, ClassVar, Sequence

import jax
import jax.numpy as jnp
from jax import grad, jacfwd
from jax.flatten_util import ravel_pytree

from .da import da_finalize, da_init, da_step
from .epoch import EpochState
from .iwls_utils import mvn_log_prob, mvn_sample, solve
from .kernel import (
    DefaultTransitionInfo,
    DefaultTuningInfo,
    ModelMixin,
    TransitionMixin,
    TransitionOutcome,
    TuningOutcome,
    WarmupOutcome,
)
from .mh import mh_step
from .pytree import register_dataclass_as_pytree
from .types import Array, KeyArray, ModelState, Position


@register_dataclass_as_pytree
@dataclass
class IWLSKernelState:
    """
    A dataclass for the state of a `IWLSKernel`, implementing the
    `liesel.goose.da.DAKernelState` protocol.
    """

    step_size: float
    error_sum: float = field(init=False)
    log_avg_step_size: float = field(init=False)
    mu: float = field(init=False)

    def __post_init__(self):
        da_init(self)


IWLSTransitionInfo = DefaultTransitionInfo
IWLSTuningInfo = DefaultTuningInfo


class IWLSKernel(ModelMixin, TransitionMixin[IWLSKernelState, IWLSTransitionInfo]):
    """
    An IWLS kernel with dual averaging and an (optional) user-defined function
    for computing the Cholesky decomposition of the Fisher information matrix,
    implementing the `liesel.goose.types.Kernel` protocol.
    """

    error_book: ClassVar[dict[int, str]] = {0: "no errors", 90: "nan acceptance prob"}
    needs_history: ClassVar[bool] = False
    identifier: str = ""

    def __init__(
        self,
        position_keys: Sequence[str],
        chol_info_fn: Callable[[ModelState], Array] | None = None,
        initial_step_size: float = 0.01,
        da_target_accept: float = 0.8,
        da_gamma: float = 0.05,
        da_kappa: float = 0.75,
        da_t0: int = 10,
    ):
        self._model = None
        self.position_keys = tuple(position_keys)
        self.chol_info_fn = chol_info_fn

        self.initial_step_size = initial_step_size

        self.da_target_accept = da_target_accept
        self.da_gamma = da_gamma
        self.da_kappa = da_kappa
        self.da_t0 = da_t0

    def _flat_log_prob_fn(
        self, model_state: ModelState, unravel_fn: Callable[[Array], Position]
    ) -> Callable[[Array], float]:
        """
        Returns a callable which takes a flat position and returns the log-probability
        of the model.
        """

        def flat_log_prob_fn(flat_position: Array) -> float:
            position = unravel_fn(flat_position)
            new_model_state = self.model.update_state(position, model_state)
            return self.model.log_prob(new_model_state)

        return flat_log_prob_fn

    def _score(
        self, model_state: ModelState, flat_score_fn: Callable[[Array], Array]
    ) -> Array:
        """
        Calls `flat_score_fn` on a flat position.

        The flat position is extracted from the `model_state`.
        """

        flat_position, _ = ravel_pytree(self.position(model_state))
        return flat_score_fn(flat_position)

    def _chol_info(
        self, model_state: ModelState, flat_hessian_fn: Callable[[Array], Array]
    ) -> Array:
        """
        Computes the Cholesky decomposition of the Fisher information matrix
        via `flat_hessian_fn`.

        The flat position is extracted from the `model_state`. If the user provided a
        `chol_info_fn` when initializing the kernel, this function is called instead.
        """

        if self.chol_info_fn is None:
            flat_position, _ = ravel_pytree(self.position(model_state))
            return jnp.linalg.cholesky(-flat_hessian_fn(flat_position))

        return self.chol_info_fn(model_state)

    def init_state(self, prng_key, model_state):
        """
        Initializes the kernel state.
        """

        return IWLSKernelState(self.initial_step_size)

    def _standard_transition(
        self,
        prng_key: KeyArray,
        kernel_state: IWLSKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[IWLSKernelState, IWLSTransitionInfo]:
        """
        Performs an MCMC transition *without* dual averaging.
        """

        key, subkey = jax.random.split(prng_key)
        step_size = kernel_state.step_size

        flat_pos, unravel_fn = ravel_pytree(self.position(model_state))
        flat_log_prob_fn = self._flat_log_prob_fn(model_state, unravel_fn)
        flat_score_fn = grad(flat_log_prob_fn)
        flat_hessian_fn = jacfwd(flat_score_fn)

        # proposal and forward probability

        score_pos = self._score(model_state, flat_score_fn)
        chol_info_pos = self._chol_info(model_state, flat_hessian_fn)
        mu_pos = flat_pos + ((step_size**2) / 2) * solve(chol_info_pos, score_pos)
        flat_prop = mvn_sample(key, mu_pos, chol_info_pos / step_size)
        proposal = unravel_fn(flat_prop)

        fwd_log_prob = mvn_log_prob(flat_prop, mu_pos, chol_info_pos / step_size)

        # backward probability

        model_state_prop = self.model.update_state(proposal, model_state)

        score_prop = self._score(model_state_prop, flat_score_fn)
        chol_info_prop = self._chol_info(model_state_prop, flat_hessian_fn)
        mu_prop = flat_prop + ((step_size**2) / 2) * solve(chol_info_prop, score_prop)
        bwd_log_prob = mvn_log_prob(flat_pos, mu_prop, chol_info_prop / step_size)

        correction = bwd_log_prob - fwd_log_prob

        info, model_state = mh_step(
            subkey, self.model, proposal, model_state, correction
        )

        return TransitionOutcome(info, kernel_state, model_state)

    def _adaptive_transition(
        self,
        prng_key: KeyArray,
        kernel_state: IWLSKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[IWLSKernelState, IWLSTransitionInfo]:
        """
        Performs an MCMC transition *with* dual averaging.
        """

        outcome = self._standard_transition(prng_key, kernel_state, model_state, epoch)

        da_step(
            outcome.kernel_state,
            outcome.info.acceptance_prob,
            epoch.time_in_epoch,
            self.da_target_accept,
            self.da_gamma,
            self.da_kappa,
            self.da_t0,
        )

        return outcome

    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: IWLSKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None = None,
    ) -> TuningOutcome[IWLSKernelState, IWLSTuningInfo]:
        """
        Currently does nothing.
        """

        info = IWLSTuningInfo(error_code=0, time=epoch.time)
        return TuningOutcome(info, kernel_state)

    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: IWLSKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> IWLSKernelState:
        """
        Resets the state of the dual averaging algorithm.
        """

        da_init(kernel_state)
        return kernel_state

    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: IWLSKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> IWLSKernelState:
        """
        Sets the step size as found by the dual averaging algorithm.
        """

        da_finalize(kernel_state)
        return kernel_state

    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: IWLSKernelState,
        model_state: ModelState,
        tuning_history: IWLSTuningInfo | None,
    ) -> WarmupOutcome[IWLSKernelState]:
        """
        Currently does nothing.
        """

        return WarmupOutcome(error_code=0, kernel_state=kernel_state)
